/*
 * buffer.c                   Copyright (C) 2004-2009 Dmitry Yakovlev
 *                                       
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "qnx4fs_fs.h"

#if defined(__QNX__) && !defined(__QNXNTO__)
#include <sys/disk.h>
#endif

static struct buffer_head *free_buffs = NULL;

struct buffer_head* chkfs_sb_bread(struct super_block* sb, u32 phys){
	struct buffer_head *bh;
	int res = -1;

	if(!free_buffs){
		bh = (struct buffer_head *)malloc(sizeof(struct buffer_head));
		if(!bh){
			perror("get read buffer");
			exit(EXIT_FAILURE);
		}
	}
	else {
		bh = free_buffs;
		free_buffs = bh->next;
	}
	bh->blknum = phys; // + sb->s_blocks_offset

#if defined(__QNX__) && !defined(__QNXNTO__)
	if(block_read(sb->s_fd, phys + sb->s_blocks_offset + 1, 1, bh->b_data) == 1)
		return bh;
#else
	if(lseek(sb->s_fd, ((off_t)phys + sb->s_blocks_offset) * QNX4FS_BLOCK_SIZE, SEEK_SET) != (off_t) -1)
	if((res = read(sb->s_fd, bh->b_data, QNX4FS_BLOCK_SIZE)) == QNX4FS_BLOCK_SIZE)
		return bh;
#endif
	chkfs_brelse(bh);
	return NULL;
}

void chkfs_brelse(struct buffer_head* bh){
	bh->next = free_buffs;
	free_buffs = bh;
}

int chkfs_bwrite(struct super_block* sb, struct buffer_head* bh){
	int res = -1;

	close(sb->s_fd);
	sb->s_fd = open(sb->s_devname, O_RDWR);

	if(sb->s_fd >= 0){
#if defined(__QNX__) && !defined(__QNXNTO__)
		res = block_write(sb->s_fd, bh->blknum + 1 + sb->s_blocks_offset, 1, bh->b_data);
		if(res > 0) res = QNX4FS_BLOCK_SIZE;
#else
		if(lseek(sb->s_fd, ((off_t)bh->blknum + sb->s_blocks_offset) * 
			QNX4FS_BLOCK_SIZE, SEEK_SET) != (off_t) -1)
			res = write(sb->s_fd, bh->b_data, QNX4FS_BLOCK_SIZE);
#endif
		close(sb->s_fd);
	}
	else {
		perror(" Cannot write to device");
	}

	sb->s_fd = open(sb->s_devname, O_RDONLY);

	if(sb->s_fd < 0){
		perror(" Cannot reopen device");
		exit(1);
	}

	return res;
}
