/*
 * super.c           Copyright (C) 2004-2009 Dmitry Yakovlev
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#if defined(__QNX__) && !defined(__QNXNTO__)
#include <sys/disk.h>
#endif

#include "qnx4fs_fs.h"

int qnx4fs_get_devsize (struct super_block *sb){
	int i = 0;

#if defined(__QNX__) && !defined(__QNXNTO__)
	char buf[QNX4FS_BLOCK_SIZE];
	struct _disk_entry de;

	if (QNX4FS_SB(sb)->s_blocks)
		goto devsize_forced;

	/*   small partition   */
	if((i = lseek(sb->s_fd, 0, SEEK_END)) != (off_t) -1) {
		i >>= QNX4FS_BLOCK_SIZE_BITS;
		if((block_read(sb->s_fd, i, 1, buf) == 1) 
			&& (block_read(sb->s_fd, i + 1, 1, buf) < 1)) {
			QNX4FS_SB(sb)->s_blocks = i;
		}
		else i = 0;
	}

	/*   large partition   */
	if(i <= 0)
	if(disk_get_entry(sb->s_fd, &de) != (-1)) {
		if((block_read(sb->s_fd, de.num_sectors, 1, buf) == 1) 
			&& (block_read(sb->s_fd, de.num_sectors + 1, 1, buf) < 1)){
			QNX4FS_SB(sb)->s_blocks = i = de.num_sectors;
		}
	}

	if(i <= 0){
		for(i = 0x7FFFFFFFL; i > 0; i >>= 1) 
			if(block_read(sb->s_fd, i, 1, buf) == 1) break;
		for(i <<= 1; i > 0; i -= 1000)
    		if(block_read(sb->s_fd, i, 1, buf) == 1) break;
		for(i += 1000; i > 0; i--)
			if(block_read(sb->s_fd, i, 1, buf) == 1) break;
		QNX4FS_SB(sb)->s_blocks = i;
		printf(" Undefined partition, %d blocks\n", i);
	}
	/* can't get perfect device size for QNX4 */
	if(i <= 0) return -1;
#else
	if (QNX4FS_SB(sb)->s_blocks)
		goto devsize_forced;

	if(lseek(sb->s_fd, 0, SEEK_END) == (off_t) -1
	|| lseek(sb->s_fd, 0, SEEK_END) == 0)
	return -1;

	QNX4FS_SB(sb)->s_blocks =
	lseek(sb->s_fd, 0, SEEK_END) >> QNX4FS_BLOCK_SIZE_BITS;
#endif

	if (sb->s_blocks_offset) {
		if (sb->s_blocks_offset > QNX4FS_SB(sb)->s_blocks) {
			fprintf(stderr," Wrong dev size/offset blocks %lu, offset %lu\n",
			QNX4FS_SB(sb)->s_blocks, sb->s_blocks_offset);
			return -1;
		}
		QNX4FS_SB(sb)->s_blocks -= sb->s_blocks_offset;
	}

devsize_forced:
	i = (((QNX4FS_SB(sb)->s_blocks + 7) >> 3) + 511) & 0xFFFFFE00;

	QNX4FS_SB(sb)->s_bitmap_new = malloc(i);
	QNX4FS_SB(sb)->s_bitmap_blocks = i >> QNX4FS_BLOCK_SIZE_BITS;

	if(QNX4FS_SB(sb)->s_bitmap_new == NULL) {
		fprintf(stderr," Unable to allocate new bitmap memory\n");
		exit(EXIT_FAILURE);
	}
	memset(QNX4FS_SB(sb)->s_bitmap_new, 0, 
	QNX4FS_SB(sb)->s_bitmap_blocks * QNX4FS_BLOCK_SIZE);
	return 0;
}

int qnx4fs_check_super (struct super_block *sbp) {
	struct buffer_head *bp;
	struct inode *ip;
	qnx4fs_dirent_t *qdp;
	unsigned long phys;
	static struct inode root_dir;
	static struct inode inodes;

	bp = sb_bread (sbp, QNX4FS_ROOT_BLOCK - 1);
	if (bp == NULL) {
		fprintf(stderr," Unable to read root block\n");
		goto out;
	}

	qdp = (qnx4fs_dirent_t *)bp->b_data;
	if (memcmp(qdp->i.name, "/", 2)) {
		fprintf(stderr," Bad root directory name\n");
		if(chkfs_press_keym(chkfs_EDIT) == chkfs_EDIT)
			chkfs_edit_dir_blk(sbp, QNX4FS_ROOT_BLOCK);
		goto outbp;
	}

	phys = le32_to_cpu(qdp->i.xt0.start);
	QNX4FS_SB(sbp)->s_bitmap_ino = direct2ino(phys, 2);
	QNX4FS_SB(sbp)->s_bitmap_blocks = phys - QNX4FS_BITMAP_BLOCK;

	brelse (bp);

	bp = sb_bread(sbp, phys - 1);
	if (bp == NULL) {
		fprintf(stderr," Unable to read root dir extent 0 block %lu\n", phys);
		if(chkfs_press_keym(chkfs_EDIT) == chkfs_EDIT)
			chkfs_edit_dir_blk(sbp, QNX4FS_ROOT_BLOCK);
		goto out;
	}

	qdp = (qnx4fs_dirent_t *)bp->b_data + 2;
	if (memcmp(qdp->i.name, ".bitmap\0", 8)) {
		fprintf(stderr," Wrong '/.bitmap' entry\n");
		if(chkfs_press_keym(chkfs_EDIT) == chkfs_EDIT)
			chkfs_edit_dir_blk(sbp, phys);
		goto outbp;
	}

	if(qnx4fs_get_devsize(sbp) < 0)
		goto outbp;

	if((le32_to_cpu(qdp->i.size) != ((QNX4FS_SB(sbp)->s_blocks) + 7) >> 3)){
		fprintf(stderr," Wrong '/.bitmap' size %u, should be %u\n",
		(unsigned)(le32_to_cpu(qdp->i.size)),  
		(unsigned)((QNX4FS_SB(sbp)->s_blocks) + 7) >> 3);
		
		if(QNX4FS_SB(sbp)->s_bitmap_blocks == 
		(le32_to_cpu(qdp->i.size) + 511) >> QNX4FS_BLOCK_SIZE_BITS){
			fprintf(stderr, "Fix bitmap size?");

			if(chkfs_press_keym (chkfs_YES) == chkfs_YES){
				qdp->i.size = cpu_to_le32(((QNX4FS_SB(sbp)->s_blocks) + 7) >> 3);
				chkfs_bwrite(sbp, bp);
				goto quitok;
			}
		}

		if(chkfs_press_keym(chkfs_EDIT) == chkfs_EDIT)
			chkfs_edit_dir_blk(sbp, phys);
		goto outbp;
	}

quitok:
	brelse(bp);

	ip = &root_dir;
	ip->i_ino = QNX4FS_ROOT_INO;
	ip->i_sb = sbp;
	qnx4fs_read_inode(ip);
	QNX4FS_SB(sbp)->s_root_dir = ip;

	ip = &inodes;
	ip->i_ino = QNX4FS_INODES_INO;
	ip->i_sb = sbp;
	qnx4fs_read_inode(ip);
	QNX4FS_SB(sbp)->s_inodes = ip;

	QNX4FS_SB(sbp)->s_bitmap_new[0] = 3; // boot & root

	if(qnx4fs_btest_and_set(sbp, (QNX4FS_SB(sbp)->s_blocks + 1),
	(QNX4FS_SB(sbp)->s_bitmap_blocks * (QNX4FS_BLOCK_SIZE << 3)) - QNX4FS_SB(sbp)->s_blocks) < 0){
		fprintf(stderr," Wrong bitmap range\n");
		return -1;
	}
	return 0;
outbp:
	brelse(bp);
out:
	return -1;
}
